module SSG.Parser.Core

import Data.Vect
import Data.String

import Control.Eff
import Derive.Prelude

%language ElabReflection

---------------------------
-- Internal parser state --
---------------------------

||| The location of a parser within a string
public export
data ParserLocation : Type where
  AtPoint : (length : Int) -> (location : Int) -> ParserLocation
  EndOfInput : (length : Int) -> ParserLocation

%runElab derive "ParserLocation" [Show]

namespace ParserLocation
  export
  fromString : String -> ParserLocation
  fromString str =
    if strLength str > 0
      then AtPoint (strLength str) 0
      else EndOfInput 0

||| The state of a parser
public export
record ParserState where
  constructor MkPS
  input : String
  location : ParserLocation

%runElab derive "ParserState" [Show]

namespace ParserState
  export
  fromString : String -> ParserState
  fromString str = MkPS str (fromString str)

  export
  toString : ParserState -> String
  toString (MkPS input (AtPoint length location)) = strSubstr location length input
  toString (MkPS input (EndOfInput length)) = ""


||| Three way result of parsing
public export
data ParseResult : Type -> Type where
  Got : t -> ParseResult t
  Failed : t -> ParseResult t
  EoF : ParseResult t

-----------------
-- Effect Type --
-----------------

||| The parser effect
export
data Parser : Type -> Type where
  -- Save and load the state
  Save : Parser ParserState
  Load : ParserState -> Parser ()
  -- Peek the next `n` characters, if there are enough, without altering the parser state
  Peek : (n : Nat) -> Parser (Maybe (Vect n Char))
  -- Parse a single char, if it satisfies a predicate
  ParseChar : (predicate : Char -> Bool) -> Parser (ParseResult Char)
  -- Parse an exact string match
  ParseExact : (str : String) -> Parser (ParseResult String)
  -- Skip a run of 0 or more chars all satisfying a predicate
  SkipChars : (predicate : Char -> Bool) -> Parser ()
  -- Detect the end of input
  ParseEoF : Parser Bool

Show (Parser a) where
  show Save = "save"
  show (Load x) = "load: \{show x}"
  show (Peek n) = "peek: \{show n}]"
  show (ParseChar predicate) = "ParseChar"
  show (ParseExact str) = "ParseExact: \{str}"
  show (SkipChars predicate) = "SkipChars"
  show ParseEoF = "ParseEof"

-----------------------
-- Effectful Actions --
-----------------------

||| Aquire a copy of the current parser state
|||
||| Intended to be used for backtracking
export
save : Has Parser fs => Eff fs ParserState
save = send Save

||| Restore the parser state to a previousy persisted copy
export
load : Has Parser fs => ParserState -> Eff fs ()
load x = send $ Load x

||| Look ahead exactly `n` characters
export
peek : Has Parser fs => (n : Nat) -> Eff fs (Maybe (Vect n Char))
peek n = send $ Peek n

||| Attempt to parse a single char satisfying the given predicate, leaving the parser
||| state unchanged on failure
export
parseChar : Has Parser fs => (predicate : Char -> Bool) -> Eff fs (ParseResult Char)
parseChar predicate = send $ ParseChar predicate

||| Parse an exact match of the given string
export
parseExact : Has Parser fs => (str : String) -> Eff fs (ParseResult String)
parseExact str = send $ ParseExact str

||| Skip a run of any number of characters satisfying the given predicate
export
skipChars : Has Parser fs => (predicate : Char -> Bool) -> Eff fs ()
skipChars predicate = send $ SkipChars predicate

||| Detect EoF
export
parseEoF : Has Parser fs => Eff fs Bool
parseEoF = send ParseEoF

--------------------
-- Effect handler --
--------------------

export
unParser : ParserState -> Parser a -> (a, ParserState)
-- Save the parser state
unParser s Save =
  (s, s)
-- Load a provided parser state
unParser s (Load x) =
  ((), x)
-- Look ahead exactly `n` characters
unParser s@(MkPS input (AtPoint length location)) (Peek n) =
  let cs = unpack . strSubstr location (cast n) $ input
  in (toVect n cs, s)
unParser s@(MkPS input (EndOfInput length)) (Peek n) =
  (Nothing, s)
-- Attempt to parse a single character matching a given predicate, bail out without
-- affecting the parser state otherwise
unParser s@(MkPS input (AtPoint length location)) (ParseChar predicate) =
  -- Saftey: We maintain location < length as a type invariant
  let c = assert_total $ strIndex input location
  in if predicate c
    then if location + 1 < length
      then (Got c, MkPS input (AtPoint length (location + 1)))
      else (Got c, MkPS input (EndOfInput length))
    else (Failed c, s)
unParser s@(MkPS input (EndOfInput length)) (ParseChar predicate) =
  (EoF, s)
-- Parse an exact string match
unParser s@(MkPS input (AtPoint length location)) (ParseExact str) =
  let substr = strSubstr location (strLength str) input
  in if substr == str
    then if location + (strLength str) < length
      then (Got substr, MkPS input (AtPoint length (location + strLength str)))
      else (Got substr, MkPS input (EndOfInput length))
    else (Failed substr, s)
unParser s@(MkPS input (EndOfInput length)) (ParseExact str) = (EoF, s)
-- Skip any number of chars satisfying a given predicate
unParser s@(MkPS input (AtPoint length location)) (SkipChars predicate) =
  case unParser s (ParseChar predicate) of
    (Got _, s) => unParser s (SkipChars predicate)
    (_, s) => ((), s)
unParser s@(MkPS input (EndOfInput length)) (SkipChars predicate) =
  ((), s)
-- "Parse" the end of input
unParser s@(MkPS input (AtPoint length location)) ParseEoF = (False, s)
unParser s@(MkPS input (EndOfInput length)) ParseEoF = (True, s)

export
runParser' : Has Parser fs =>
  (s : ParserState) -> Eff fs t -> Eff (fs - Parser) (t, ParserState)
runParser' s =
  handleRelayS s (\state, t => pure (t, state)) $ \s2, ps, f =>
    let (a, st) = unParser s2 ps
    in f st a

export
runParser : Has Parser fs =>
  (s : String) -> Eff fs t -> Eff (fs - Parser) (t, String)
runParser s x =
  map (\(x,y) => (x, toString y)) $ runParser' (fromString s) x
